local L = LibStub("AceLocale-3.0"):GetLocale("GuildNoteSetter", true);
local blCommPrefix = "GBlackList";
local BLACK_LIST_COLUMN_INFO = {
	[1] = {
		title = L["Name"],
		width = 100,
		attribute = "name",
	},

	-- [2] = {
	-- 	title = L["Added by"],
	-- 	width = 100,
	-- 	attribute = "addedBy",
	-- },

	-- [3] = {
	-- 	title = L["Date"],
	-- 	width = 120,
	-- 	attribute = "date",
	-- },

	[2] = {
		title = L["Reason"],
		width = 0,
		attribute = "reason",
	},
};

function GuildNoteSetter:CreateBlackListUI()
	self.db.realm[self.guildName] = self.db.realm[self.guildName] or {};
	self.db.realm[self.guildName].blackList = self.db.realm[self.guildName].blackList or { list = {}, updateTime = 0 };
	self.blackList = self.db.realm[self.guildName].blackList;

	-- Fill self.blackListNew from self.blackList
	self.blackListNew = self:table_deep_copy(self.blackList);

	if (not IsAddOnLoaded("Blizzard_Communities") and not LoadAddOn("Blizzard_Communities")) then
		return;
	end

	-- Add black list tab in communities frame
	local GuildBlackListButton = CreateFrame("Button", "GuildBlackListButton", CommunitiesFrame, "GuildBlackListButtonTemplate");
	local GuildBlackListFrame = CreateFrame("Frame", "GuildBlackListFrame", CommunitiesFrame, "GuildBlackListFrameTemplate");
	-- local GuildBlackListEditFrame = CreateFrame("Frame", "GuildBlackListEditFrame", CommunitiesFrame, "GuildBlackListEditFrameTemplate");

	hooksecurefunc(CommunitiesFrame, "SetDisplayMode", self.SetDisplayMode_override);
end

SafeStaticPopupDialogs["GNS_BLACKLIST_REMOVE_CONFIRM"] = {
	text = L["Are you sure you want to remove the character %s from the blacklist?"],
	button1 = L["Yes"],
	button2 = L["No"],
	OnAccept = function(frame, data)
		local playerName = data;
		local gblf = GuildBlackListFrame;

		if (gblf.GuildBlackListEditFrame:IsShown() and gblf.GuildBlackListEditFrame.playerName == playerName) then
			gblf.GuildBlackListEditFrame.playerName = "";
			gblf.GuildBlackListEditFrame:Hide();
		end

		GuildNoteSetter:BlacklistRemoveRecord(playerName);
		gblf.GuildBlackListApplyButton:Enable();
		gblf.GuildBlackListRevertButton:Enable();
		gblf:UpdateMemberList();
	end,
	timeout = 0,
	whileDead = true,
	hideOnEscape = true,
	preferredIndex = 3,  -- avoid some UI taint, see http://www.wowace.com/announcements/how-to-avoid-some-ui-taint/
};

function GuildNoteSetter:SetDisplayMode_override(displayMode)
	if displayMode == COMMUNITIES_FRAME_DISPLAY_MODES.GUILD_INFO then
		GuildBlackListButton:SetShown(true);
	else
		GuildBlackListButton:SetShown(false);
	end
end

function GuildBlackListButton_OnLoad(self)
	self:SetText(L["Blacklist"]);
end

function GuildBlackListButton_OnClick(self)
	local gblf = GuildBlackListFrame
	if not gblf:IsShown() then
		gblf:Show();
	else
		gblf:Hide();
	end
end

function GuildBlackListButton_OnHide(self)
	GuildBlackListFrame:Hide();
end

-- function GuildNoteSetter:BlackListButtonTooltipShow(frame, text)
-- 	GameTooltip:SetOwner(frame, "ANCHOR_RIGHT");
-- 	GameTooltip:SetText(L[text]);
-- 	GameTooltip:Show();
-- end
-- /run print(issecurevariable(CommunitiesFrame, "displayMode"))
-------------------------------------------------------------------------------------------------
-- BlackList frame
-------------------------------------------------------------------------------------------------
GuildBlackListMixin = {};

function GuildBlackListMixin:RefreshListDisplay()
	local scrollFrame = self.ListScrollFrame;
	local offset = HybridScrollFrame_GetOffset(scrollFrame);
	local buttons = scrollFrame.buttons;

	local usedHeight = 0;
	local height = buttons[1]:GetHeight();
	local memberList = self.memberList or {};
	local orderedIndex = self.orderedIndex or {};
	for i = 1, #buttons do
		local displayIndex = i + offset;
		local button = buttons[i];
		if displayIndex <= #orderedIndex then
			local name = orderedIndex[displayIndex];
			local memberInfo = GuildNoteSetter:table_deep_copy(memberList[name]);
			memberInfo.name = name;
            button:SetMember(memberInfo);
            button:Show();
            usedHeight = usedHeight + height;
		else
            button:SetMember(nil);
            button:Hide();
		end
	end

	local totalNum = #memberList;
	if displayInvitations then
		totalNum = totalNum + #invitations + 2;
	end

	HybridScrollFrame_Update(scrollFrame, height * totalNum, usedHeight);
end

function GuildBlackListMixin:UpdateMemberList()
	self.memberList = GuildNoteSetter:table_deep_copy(GuildNoteSetter.blackListNew.list);
	self:SortByName();
	self:Update();
end

function GuildBlackListMixin:Update()
	self.ColumnDisplay.Date:SetText(date(L["Updated"] .. " %Y-%m-%d %H:%M", GuildNoteSetter.blackListNew.updateTime));
	self:RefreshLayout();
	self:RefreshListDisplay();
end

function GuildBlackListMixin:MarkSortDirty()
	self.sortDirty = true;
end

function GuildBlackListMixin:MarkMemberListDirty()
	self.memberListDirty = true;
end

function GuildBlackListMixin:IsSortDirty()
	return self.sortDirty;
end

function GuildBlackListMixin:IsMemberListDirty()
	return self.memberListDirty;
end

function GuildBlackListMixin:ClearSortDirty()
	self.sortDirty = nil;
end

function GuildBlackListMixin:ClearMemberListDirty()
	self.memberListDirty = nil;
end

function GuildBlackListMixin:SortList()
    self:SortByName();
	self:RefreshListDisplay();
end

function GuildBlackListMixin:OnLoad()
	self:RegisterForDrag("LeftButton");
	self.TitleText:SetText(L["Blacklist"]);
	self.TitleText:SetMaxLines(1);
	self.TitleText:SetHeight(20);
	self.ListScrollFrame.update = function()
		self:Update();
	end;

	self.ListScrollFrame.scrollBar.doNotHide = true;
	self.ListScrollFrame.scrollBar:SetValue(0);

	self:UpdateMemberList();

	self:RefreshLayout();
	self:RefreshListDisplay();
end

function GuildBlackListMixin:OnDragStart()
    if ( not self.isMoving ) and ( self.isLocked ~= 1 ) then
        self:StartMoving();
        self.isMoving = true;
    end
end

function  GuildBlackListMixin:OnDragStop(e)
    if ( self.isMoving ) then
        self:StopMovingOrSizing();
        self.isMoving = false;
    end
end

function GuildBlackListMixin:OnShow()
	PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON);
	if not GuildNoteSetter:CanEditBlackList() then 
		self.GuildBlackListAddButton:Disable();
	else
		self.GuildBlackListAddButton:Enable();
	end
	self:UpdateMemberList();
end

function GuildBlackListMixin:OnHide()
	PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_OFF);
	GuildBlackListFrame.GuildBlackListEditFrame.playerName = "";
	GuildBlackListFrame.GuildBlackListEditFrame:Hide();
	if SafeStaticPopup_Visible("GNS_BLACKLIST_REMOVE_CONFIRM") then
		SafeStaticPopup_Hide("GNS_BLACKLIST_REMOVE_CONFIRM");
	end
end

function GuildBlackListMixin:OnUpdate()
	if self:IsMemberListDirty() then
		self:UpdateMemberList();
		self:ClearMemberListDirty();
	end

	if self:IsSortDirty() then
		self:SortList();
		self:ClearSortDirty();
		self:UpdateMemberCount();
	end
end

function GuildBlackListMixin:RefreshLayout()
	if not self.ListScrollFrame.buttons then
		HybridScrollFrame_CreateButtons(self.ListScrollFrame, "GuildBlackListEntryTemplate", 0, 0);
	end

	self.ColumnDisplay:Hide();
	self.columnInfo = BLACK_LIST_COLUMN_INFO;
	self.ColumnDisplay:LayoutColumns(BLACK_LIST_COLUMN_INFO);
	self.ColumnDisplay:Show();
end

function GuildBlackListMixin:GetCommunitiesFrame()
	return self:GetParent();
end

function GuildBlackListMixin:OnClubMemberButtonClicked(entry, button)
	if not GuildNoteSetter:CanEditBlackList() then
		return;
	end
	if button == "RightButton" then
		self.selectedEntry = entry;
		ToggleSafeDropDownMenu(1, nil, self.DropDown, entry, 0, 0);
		return;
	end
end

function GuildBlackListMixin:GetSelectedEntry()
	return self.selectedEntry;
end

function GuildBlackListMixin:OnDropDownClosed()
	self.selectedEntry = nil;
end

function GuildBlackListMixin:GetSelectedClubId()
	return self:GetCommunitiesFrame():GetSelectedClubId();
end

function GuildBlackListMixin:GetSelectedClubInfo()
	return self:GetCommunitiesFrame():GetSelectedClubInfo();
end

function GuildBlackListMixin:SortByName()
	local orderedIndex = {};
	for key in pairs(self.memberList) do
		table.insert(orderedIndex, key);
	end
	table.sort(orderedIndex);
	self.orderedIndex = orderedIndex;
end

-------------------------------------------------------------------------------------------------
-- BlackList entry
-------------------------------------------------------------------------------------------------
GuildBlackListEntryMixin = {};

function GuildBlackListEntryMixin:GetMemberList()
	return self:GetParent():GetParent():GetParent();
end

function GuildBlackListEntryMixin:SetMember(memberInfo)
	if memberInfo then
		self.memberInfo = memberInfo;
	else
		self.memberInfo = nil;
	end

	self:Update();
end

function GuildBlackListEntryMixin:GetMemberInfo()
	return self.memberInfo;
end

function GuildBlackListEntryMixin:GetMemberInfo()
	return self.memberInfo;
end

function GuildBlackListEntryMixin:GetMemberId()
	return self.memberInfo and self.memberInfo.memberId or nil;
end

function GuildBlackListEntryMixin:OnClick(button)
    self:GetMemberList():OnClubMemberButtonClicked(self, button);
end

function GuildBlackListEntryMixin:Update()
	local memberInfo = self:GetMemberInfo();
    local hasMemberInfo = memberInfo ~= nil;
    self.Name:SetShown(hasMemberInfo);
    -- self.AddedBy:SetShown(hasMemberInfo);
    -- self.Date:SetShown(hasMemberInfo);
    self.Reason:SetShown(hasMemberInfo);
	if not hasMemberInfo then
		return;
    end
    
    self.Name:SetText(memberInfo.name or "")
    -- self.AddedBy:SetText(memberInfo.addedBy or "")
    -- self.Date:SetText(date("%Y-%m-%d %H:%M", memberInfo.since) or "")
    self.Reason:SetText(memberInfo.reason or "")
end

function GuildBlackListEntryMixin:OnEnter()
	local memberInfo = self:GetMemberInfo();
	if memberInfo then
		GameTooltip:SetOwner(self);
		GameTooltip:AddLine(memberInfo.name);

		GameTooltip:AddLine("|n|cffffd100" .. L["Added by:"] .. "|r " .. memberInfo.addedBy .. "|n" .. date("%Y-%m-%d %H:%M", memberInfo.since), 1.0, 1.0, 1.0);
		GameTooltip:AddLine("|n|cffffd100" .. L["Reason:"], 1.0, 1.0, 1.0);
		GameTooltip:AddLine(memberInfo.reason, 1.0, 1.0, 1.0, true);

		GameTooltip:Show();
	end
end

function GuildBlackListEntryMixin:OnLeave()
	GameTooltip:Hide();
end

-------------------------------------------------------------------------------------------------
-- BlackList dropdown menu
-------------------------------------------------------------------------------------------------

function GuildBlackListDropDown_OnLoad(self)
	SafeUIDropDownMenu_Initialize(self, GuildBlackListDropDown_Initialize, "MENU");
end

function GuildBlackListDropDown_OnHide(self)
	local GuildBlackList = self:GetParent();
	GuildBlackList:OnDropDownClosed();
end

function GuildBlackListDropDown_Initialize(self, level)
	local GuildBackList = self:GetParent();
	local SelectedGuildBackListEntry = GuildBackList:GetSelectedEntry();
	if not SelectedGuildBackListEntry then
		return;
	end
	
	local memberInfo = SelectedGuildBackListEntry:GetMemberInfo();
	local entry = SafeUIDropDownMenu_CreateInfo();
	if memberInfo and memberInfo.name then
		local playerName = memberInfo.name;
		-- Add title
		entry.text = playerName;
		entry.isTitle = true;
		entry.notCheckable = true;
		SafeUIDropDownMenu_AddButton(entry);

		-- Edit line
		entry.isTitle = false;
		entry.disabled = false;
		entry.text = L["Edit"];
		entry.func = function(frame,_,_,checked)
			PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON);
			GuildNoteSetter:BlackListEditDetailsFor(playerName);
		end
		SafeUIDropDownMenu_AddButton(entry);

		-- Remove line
		entry.text = L["Remove"];
		entry.func = function(frame,_,_,checked)
			PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON);
			SafeStaticPopup_Show("GNS_BLACKLIST_REMOVE_CONFIRM", "|cffffd200" .. playerName .. "|r", "", playerName)
		end
		SafeUIDropDownMenu_AddButton(entry);
	end
end

-------------------------------------------------------------------------------------------------
-- BlackList buttons
-------------------------------------------------------------------------------------------------
function GuildBlackListAddButton_OnLoad(self)
	self:SetText(L["Add"]);
	self.tooltipText = L["Add a player to the black list.|n|nShift-Click for batch adding of multiple players at once."];
end

function GuildBlackListAddButton_OnClick(self)
	PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON);
	GuildNoteSetter:BlackListEditDetailsFor("");
	-- TODO VVV add impl. of mass add
end

function GuildBlackListApplyButton_OnLoad(self)
	self:SetText(L["Apply"]);
	self.tooltipText = L["Apply all blacklist changes and send update to the guildmates."];
	self:Disable();
end

function GuildBlackListApplyButton_OnClick(self)
	PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON);
	GuildNoteSetter:ApplyBlackList();
	GuildBlackListFrame.GuildBlackListApplyButton:Disable();
	GuildBlackListFrame.GuildBlackListRevertButton:Disable();
	GuildBlackListFrame:UpdateMemberList();
	GuildNoteSetter:ScheduleNonCombatTimer("DetectBlackListedPlayers", 3);
end

function GuildBlackListRevertButton_OnLoad(self)
	self:SetText(L["Revert"]);
	self.tooltipText = L["Revert all blacklist changes."];
	self:Disable();
end

function GuildBlackListRevertButton_OnClick(self)
	PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON);
	GuildNoteSetter:RevertBlackList();
	GuildBlackListFrame.GuildBlackListApplyButton:Disable();
	GuildBlackListFrame.GuildBlackListRevertButton:Disable();
	GuildBlackListFrame:UpdateMemberList();
end

-------------------------------------------------------------------------------------------------
-- BlackList details frame
-------------------------------------------------------------------------------------------------
function GuildNoteSetter:BlackListEditFrame_OnLoad(frame)
	GuildBlackListEditOkBtn:SetText(L["Save"]);
	GuildBlackListEditCancelBtn:SetText(L["Cancel"]);

	GuildBlackListEditNameLabel:SetText(L["Name:"]);
	GuildBlackListEditName.Instructions:SetText(L["Enter the character name"]);

	GuildBlackListEditReasonLabel:SetText(L["Reason:"]);
	GuildBlackListEditReason.instructions = L["Details why the character was black-listed."];
	GuildBlackListEditReason.EditBox.Instructions:SetText(GuildBlackListEditReason.instructions);
end

function GuildNoteSetter:BlackListEditFrame_OnShow(frame)
	-- InputScrollFrame_OnLoad(GuildBlackListEditReason);	-- This is necessary to update the internal edit-box width  -- Looks like everything is fine now and can be deleted

	GuildBlackListEditName:SetText(frame.playerName);

	local reason
	if frame.playerName ~= "" then
		-- Edit character that is already on the list
		reason = self.blackListNew.list[frame.playerName].reason

		GuildBlackListEditName:Disable()
		GuildBlackListEditName.Left:Hide();
		GuildBlackListEditName.Right:Hide();
		GuildBlackListEditName.Middle:Hide();
	else
		-- Add new character
		reason = ""

		GuildBlackListEditName:Enable()
		GuildBlackListEditName.Left:Show();
		GuildBlackListEditName.Right:Show();
		GuildBlackListEditName.Middle:Show();
	end
	GuildBlackListEditReason.EditBox:SetText(reason);
end

-- Empty name means "new"
function GuildNoteSetter:BlackListEditDetailsFor(name)
	local gblf = GuildBlackListFrame;
	if gblf.GuildBlackListEditFrame:IsShown() and gblf.GuildBlackListEditFrame.playerName == name then
		return;
	end

	gblf.GuildBlackListEditFrame.playerName = name;
	if gblf.GuildBlackListEditFrame:IsShown() then
		self:BlackListEditFrame_OnShow(gblf.GuildBlackListEditFrame);
	else
		gblf.GuildBlackListEditFrame:Show();
	end
end

function GuildBlackListEditOkBtn_OnLoad(self)
	self.tooltipText = L["Save changes and close this window"];
end

function GuildNoteSetter:BlackListEditOkBtn_OnClick(frame)
	PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON);

	-- Verify the reason
	local reason = strtrim(GuildBlackListEditReason.EditBox:GetText());
	if (not reason or reason == "") then
		self:PopUIError(L["You must specify a reason for black-listing"]);
		return;
	end

	-- Verify the name (non-empty, not yet in the list)
	local newName = strtrim(GuildBlackListEditName:GetText());
	local gblf = GuildBlackListFrame
	if gblf.GuildBlackListEditFrame.playerName == "" then
		if (not newName or newName == "") then
			self:PopUIError(L["The character name is not specified"]);
			return;
		end

		if (self.blackListNew.list[newName]) then
			self:PopUIError(L["The character is already on the black list"]);
			return;
		end
	end

	if gblf.GuildBlackListEditFrame.playerName == "" then
		self:BlacklistAddRecord(newName, reason);
	else
		self:BlacklistUpdateRecord(gblf.GuildBlackListEditFrame.playerName, reason);
	end

	gblf.GuildBlackListEditFrame.playerName = "";
	gblf.GuildBlackListEditFrame:Hide();

	gblf.GuildBlackListApplyButton:Enable();
	gblf.GuildBlackListRevertButton:Enable();
	gblf:UpdateMemberList();
end

function GuildBlackListEditCancelBtn_OnLoad(self)
	self.tooltipText = L["Close this window without saving"];
end

function GuildNoteSetter:BlackListEditCancelBtn_OnClick(frame)
	PlaySound(SOUNDKIT.IG_MAINMENU_OPTION_CHECKBOX_ON);

	GuildBlackListFrame.GuildBlackListEditFrame.playerName = "";
	GuildBlackListFrame.GuildBlackListEditFrame:Hide();
end

function GuildNoteSetter:PopUIError(text)
	UIErrorsFrame:AddMessage(text, 1.0, 0.1, 0.1, 1.0);
	PlaySound(SOUNDKIT.RAID_WARNING);
end

-------------------------------------------------------------------------------------------------
-- BlackList logic stuff
-------------------------------------------------------------------------------------------------
function GuildNoteSetter:OrderedBlacklistIndex(list)
    local orderedIndex = {}
    for key in pairs(list) do
        table.insert(orderedIndex, key)
    end
    table.sort(orderedIndex)
    return orderedIndex
end

function GuildNoteSetter:RevertBlackList()
	-- Just fill self.blackListNew from self.blackList
	self:table_deep_copy(self.blackList, self.blackListNew)

	self:Print(L["The blacklist was reverted."]);
end

function GuildNoteSetter:ApplyBlackList()
	if not self:CanEditBlackList() then return end

	local added, removed, changed = self:CountBlackListChanges()
	if added > 0 or removed > 0 or changed > 0 then
		local str = L["Blacklist changes approved: "]

		if added > 0 then str = str .. added .. " " .. L["players added."] .. " " end
		if removed > 0 then str = str .. removed .. " " .. L["players removed."] end
		if changed > 0 then str = str .. changed .. " " .. L["players edited."] end
		self:Print(str);

		-- Fill self.blackList from self.blackListNew
		self.blackListNew.updateTime = GetServerTime()
		self:table_deep_copy(self.blackListNew, self.blackList)

		self.blOutdated = false
		self:TransmitBlackListMsg(blMsgAnnounce, self.blackList.updateTime, "GUILD");
	else
		self:Print(L["The blacklist hasn't changed."]);
	end
end

function GuildNoteSetter:BlacklistAddRecord(name, newReason)
	self.blackListNew.list[name] = {
		reason = newReason,
		addedBy = UnitName("player"),
		since = GetServerTime(),
	};
end

function GuildNoteSetter:BlacklistRemoveRecord(name)
	self.blackListNew.list[name] = nil;
end

function GuildNoteSetter:BlacklistUpdateRecord(name, newReason)
	self.blackListNew.list[name].reason = newReason;
end

function GuildNoteSetter:CanEditBlackList()
	return self.roster:CanPlayerEditBL(self.playerName);
end

function GuildNoteSetter:IsTrustedBlackListSrc(name)
	return self.roster:IsPlayerTrustedBLSrc(name);
end

function GuildNoteSetter:IsMeTrustedBlackListSrc()
	return self:IsTrustedBlackListSrc(self.playerName);
end

function GuildNoteSetter:CountBlackListChanges()
	local added = 0
	local removed = 0
	local changed = 0

	local newList = self.blackListNew.list
	local oldList = self.blackList.list

	for name in pairs(newList) do
		if oldList[name] then
			if oldList[name].reason ~= newList[name].reason then
				changed = changed + 1
			end
		else
			added = added + 1
		end
	end

	for name in pairs(oldList) do
		if not newList[name] then
			removed = removed + 1
		end
	end

	return added, removed, changed
end

function GuildNoteSetter:DetectBlackListedPlayers()
	if self.guildName == nil then
		return;
	end
	for name in pairs(self.blackList.list) do
		if (not self.detectedPlayers[name]) and self.roster:PlayerExists(self:LongName(name)) then
			self.detectedPlayers[name] = 1
			self:Printf("|cffff1919" .. L["ATTENTION: Black-listed character found in the guild"] .. "|r - " .. name)
		end
	end

--	if CanGuildRemove() then
--		self:KickNextBlackListedPlayer();
--	end
end

--function GuildNoteSetter:KickNextBlackListedPlayer()
--	for name in pairs(self.detectedPlayers) do
--		if self.detectedPlayers[name] == 1 then
--			self.detectedPlayers[name] = 2
--			StaticPopup_Show("GNS_KICK_BLACKLISTED_CONFIRM", "|cffffd200" .. name .. "|r", "", name)
--			return
--		end
--	end
--end

--StaticPopupDialogs["GNS_KICK_BLACKLISTED_CONFIRM"] = {
--	text = L["Black-listed character %s is found in the guild. Kick it?"],
--	button1 = L["Yes"],
--	button2 = L["No"],
--	OnAccept = function(frame, data)
--		local playerName = data;
--		
--		GuildUninvite(playerName);
--		GuildNoteSetter:ScheduleNonCombatTimer("KickNextBlackListedPlayer", 0.1)
--	end,
--	OnCancel = function()
--		GuildNoteSetter:ScheduleNonCombatTimer("KickNextBlackListedPlayer", 0.1)
--	end,
--	timeout = 0,
--	whileDead = true,
--	hideOnEscape = true,
--	preferredIndex = 3,  -- avoid some UI taint, see http://www.wowace.com/announcements/how-to-avoid-some-ui-taint/
--}

-------------------------------------------------------------------------------------------------
-- BlackList P2P transfer functions
--
-- Protocol logic:
--  1. On login wait 10sec, then announce our DB version (the first announce is important). Keep reannouncing
--     every 5 minutes. However, if the same DB version was announced by someone else in the last 5 sec,
--     postpone our announce for a random interval of 5-60 sec.
--  2. If someone announced a newer version, stop announcing ours.
--  3. If someone announced version newer than ours, wait random period of 1-60 seconds and request data
--     transfer. However, if any others announced same version during this wait period - request from random
--     player in the list.
--  4. If someone announced version older than ours, wait random period of 1-60 seconds and inform them that
--     we have newer.
--  5. Limit number of simultaneous data transfers to 3. If data request is received when these transfers are
--     in progress, ask the peer to reapply after 10-30 seconds.
-------------------------------------------------------------------------------------------------
local blMsgAnnounce = "DBVersion";
local blMsgUpdateReq = "DBReq";
local blMsgDBUpdate = "DBUpdate";
local blMsgBusy = "Busy";
local blMsgVerReq = "DBVerReq";

local blAnnouncePeriod_s = 5 * 60;
local blMinAnnounceInterval_s = 5;
local blAnnouncePostponeRnd_s = 60;
local blMinRequestDelay_s = 1;
local blRequestRnd_s = 60;
local blRequestTimeout_s = 20;
local blPollRnd_s = 10;
local blMaxConcurrentOutTransfers = 3;

function GuildNoteSetter:OnBlackListCommReceived(prefix, msg, distribution, sender)
	if prefix ~= blCommPrefix then return end

	-- In combat we just process guild announces, ignore all whisper communications.
	if InCombatLockdown() and distribution ~= "GUILD" then return end

	sender = self:LongName(sender);
	if sender == self.playerName then return end		-- Ignore messages from ourselves

	if distribution == "GUILD" then
		self:OnBlackListGuildMsg(msg, sender);
	elseif distribution == "WHISPER" then
		self:OnBlackListWhisperMsg(msg, sender);
	end
end

function GuildNoteSetter:OnBlackListGuildMsg(encodedMsg, sender)
	if string.len(encodedMsg) > 64 then return end				-- Ignore long guild messages, as these are parsed even in combat
	-- if not self:IsTrustedBlackListSrc(sender) then return end	-- Only some players are trusted

	local msg = self:EncodedStringToTable(encodedMsg);
	if not type(msg) == "table" then
		self:Debug("Decoding failed. " .. tostring(msg));
		return
	end

	if msg.m == blMsgAnnounce then
		local remoteUpdateTime = msg.d
		if remoteUpdateTime > self.blackList.updateTime then
			self:OnBlackListRemoteHasNewer(sender);
		else
			if (not InCombatLockdown()) and (remoteUpdateTime < self.blackList.updateTime) and self:IsMeTrustedBlackListSrc() then
				local informAfter_s = blMinRequestDelay_s + blRequestRnd_s * fastrandom();
				self:ScheduleNonCombatTimer("BlackListInformRemoteWeHaveUpdate", informAfter_s, sender)
			end

			self.blLastAnnounceTime = GetTime();
		end
	elseif msg.m == blMsgVerReq then
		local informAfter_s = blMinRequestDelay_s + blPollRnd_s * fastrandom();
		self:ScheduleNonCombatTimer("BlackListInformRemoteWeHaveUpdate", informAfter_s, sender)
	end
end

function GuildNoteSetter:AnnounceOurBlackListDb()
    self:Debug("aaa");
	if not self:IsMeTrustedBlackListSrc() then return end
    self:Debug("bbb");

	local timeSinceLastAnnounce = GetTime() - self.blLastAnnounceTime;
	local nextAnnounceAfter_s = blAnnouncePeriod_s
	if timeSinceLastAnnounce > blMinAnnounceInterval_s then
		if not self.blOutdated then
			self:Debug("Announcing our blacklist DB.");
			self:TransmitBlackListMsg(blMsgAnnounce, self.blackList.updateTime, "GUILD");
		end
	else
		nextAnnounceAfter_s = blMinAnnounceInterval_s + blAnnouncePostponeRnd_s * fastrandom();
	end

	-- Schedule next announce
	self:ScheduleNonCombatTimer("AnnounceOurBlackListDb", nextAnnounceAfter_s)
end

function GuildNoteSetter:OnBlackListRemoteHasNewer(sender)
	if self.blInTransferFrom then return end			-- We already in progress of receiving a new one

	self:Debug("Our blacklist is outdated,", sender, "has a newer version.");

	self.blOutdated = true;

	table.insert(self.blSourcesForUpdate, sender);
	if 1 == #self.blSourcesForUpdate then	-- This is the first, start timer
		local requestAfter_s = blMinRequestDelay_s + blRequestRnd_s * fastrandom();
		self:ScheduleNonCombatTimer("BlackListRequestRemoteUpdate", requestAfter_s)
	end
end

function GuildNoteSetter:TransmitBlackListMsg(msgType, data, channel, target, prio, callback)
	local transmit = {
		m = msgType,
		d = data,
	};
	local encoded = self:TableToEncodedString(transmit);
	self:SendCommMessage(blCommPrefix, encoded, channel, target, prio, callback);
end

function GuildNoteSetter:OnBlackListWhisperMsg(encodedMsg, sender)
	local msg = self:EncodedStringToTable(encodedMsg);
	if not type(msg) == "table" then
		self:Debug("Decoding failed. " .. tostring(msg));
		return
	end

	if msg.m == blMsgUpdateReq then
		self:OnBlackListRemoteUpdateReq(sender);
	elseif msg.m == blMsgDBUpdate then
		self:OnBlackListRemoteUpdateReceived(sender, msg.d)
	elseif msg.m == blMsgAnnounce then
		local remoteUpdateTime = msg.d
		self.blPeerVers[sender] = remoteUpdateTime;
		if remoteUpdateTime > self.blackList.updateTime and self:IsTrustedBlackListSrc(sender) then
			self:OnBlackListRemoteHasNewer(sender);
		end
	elseif msg.m == blMsgBusy then
		self:OnBlackListRemoteBusy(sender)		-- The remote is too busy, retry later
	end
end

function GuildNoteSetter:BlackListRequestRemoteUpdate()
	-- Choose random source
	local srcName = self.blSourcesForUpdate[fastrandom(#self.blSourcesForUpdate)]
	self.blInTransferFrom = srcName
	self.blSourcesForUpdate = {}

	self:Debug("Requesting blacklist update from", srcName);
	self:TransmitBlackListMsg(blMsgUpdateReq, nil, "WHISPER", srcName);

	self.blReqTimeoutTimer = self:ScheduleTimer("BlackListRemoteRequestTimeout", blRequestTimeout_s)
end

function GuildNoteSetter:BlackListRemoteRequestTimeout()
	self:Debug("Blacklist request from", self.blInTransferFrom, "timed out.");

	self.blInTransferFrom = nil
	self.blReqTimeoutTimer = nil
end

function GuildNoteSetter:OnBlackListRemoteUpdateReq(sender)
	self:Debug("Blacklist update request from", sender);

	if (self.blOutTransfers >= blMaxConcurrentOutTransfers) then
		self:Debug("Too many transfers in progress.");
		self:TransmitBlackListMsg(blMsgBusy, nil, "WHISPER", sender);
	else
		self:Debug("Sending blacklist to", sender);

		self.blOutTransfers = self.blOutTransfers + 1
		self:TransmitBlackListMsg(blMsgDBUpdate, self.blackList, "WHISPER", sender, "BULK", function(arg, sent, total)
			if (sent >= total) then
				self:Debug("Sending done.");
				self.blOutTransfers = self.blOutTransfers - 1
			end
		end);
	end
end

function GuildNoteSetter:OnBlackListRemoteUpdateReceived(sender, data)
	if sender ~= self.blInTransferFrom then return end		-- Impostor? Only receive from the one we requested

	-- Basic verification of the received data
	if (type(data.updateTime) ~= "number") or (data.updateTime <= self.blackList.updateTime) or type(data.list) ~= "table" then
		return
	end

	self:table_deep_copy(data, self.blackList)
	--self:table_deep_copy(data, self.blackListNew)

	self.blInTransferFrom = nil
	self.blOutdated = false

	-- Cancel the timeout timer
	self:CancelTimer(self.blReqTimeoutTimer)
	self.blReqTimeoutTimer = nil

	self:Printf(L["Blacklist update received from %s"], sender);

	GuildBlackListFrame:UpdateMemberList();
	self:ScheduleNonCombatTimer("DetectBlackListedPlayers", 3)
end

function GuildNoteSetter:OnBlackListRemoteBusy(sender)
	if sender ~= self.blInTransferFrom then return end		-- Impostor? Only receive from the one we requested

	self:Debug(sender, "is busy and can't send data.");

	self.blInTransferFrom = nil

	-- Cancel the timeout timer
	self:CancelTimer(self.blReqTimeoutTimer)
	self.blReqTimeoutTimer = nil

	-- Schedule a new attempt later
	self:OnBlackListRemoteHasNewer(sender);
end

function GuildNoteSetter:BlackListInformRemoteWeHaveUpdate(remoteName)
	self:Debug(remoteName, "has outdated blacklist. Inform we have newer.");

	self:TransmitBlackListMsg(blMsgAnnounce, self.blackList.updateTime, "WHISPER", remoteName);
end

function GuildNoteSetter:BlacklistPollDbVersions()
	self:Print(L["Polling guild mates, please wait..."]);

	self.blPeerVers = {};
	self:TransmitBlackListMsg(blMsgVerReq, nil, "GUILD");
	self:ScheduleNonCombatTimer("BlacklistVersionsPollEnded", blPollRnd_s + 3)
end

function GuildNoteSetter:BlacklistVersionsPollEnded()
	local names = {}
	for n in pairs(self.blPeerVers) do
		table.insert(names, n)
	end
	table.sort(names)

	for i, n in ipairs(names) do
		local s = "";
		if self:IsTrustedBlackListSrc(n) then s = "*" end

		s = s .. self:ShortName(n) .. " (" .. date("%Y-%m-%d %H:%M", self.blPeerVers[n]) .. ")"
		self:Print(s);
	end

	self:Printf(L["Found %d guild mates with the addon installed:"], table.getn(names));
end
