-- This is an exact replica of Blizzard StaticPopup code, only all names replaced to SafeStaticPopup
-- It enables us to use this functionality without bumping into taint issues.

SafeStaticPopup_DisplayedFrames = { };

SAFESTATICPOPUP_NUMDIALOGS = 4;
SAFESTATICPOPUP_TIMEOUT = 60;
SAFESTATICPOPUP_TEXTURE_ALERT = "Interface\\DialogFrame\\UI-Dialog-Icon-AlertNew";
SAFESTATICPOPUP_TEXTURE_ALERTGEAR = "Interface\\DialogFrame\\UI-Dialog-Icon-AlertOther";
SafeStaticPopupDialogs = { };


-- local function SetupLockOnDeclineButtonAndEscape(self, declineTimeLeft)
-- 	self.declineTimeLeft = declineTimeLeft or .5;
-- 	self.button2:SetButtonState("NORMAL", true);
-- 	self.ticker = C_Timer.NewTicker(.5, function()
-- 		self.declineTimeLeft = self.declineTimeLeft - .5;
-- 		if (self.declineTimeLeft == 0) then
-- 			self.ticker:Cancel();
-- 			self.button2:SetButtonState("NORMAL", false);
-- 			return;
-- 		else
-- 			self.button2:SetButtonState("NORMAL", true);
-- 		end
-- 	end);
-- 	self.hideOnEscape = false;
-- end


function SafeStaticPopup_FindVisible(which, data)
	local info = SafeStaticPopupDialogs[which];
	if ( not info ) then
		return nil;
	end
	for index = 1, SAFESTATICPOPUP_NUMDIALOGS, 1 do
		local frame = _G["SafeStaticPopup"..index];
		if ( frame:IsShown() and (frame.which == which) and (not info.multiple or (frame.data == data)) ) then
			return frame;
		end
	end
	return nil;
end

function SafeStaticPopup_Visible(which)
	for index = 1, SAFESTATICPOPUP_NUMDIALOGS, 1 do
		local frame = _G["SafeStaticPopup"..index];
		if( frame:IsShown() and (frame.which == which) ) then
			return frame:GetName(), frame;
		end
	end
	return nil;
end

function SafeStaticPopup_Resize(dialog, which)
	local info = SafeStaticPopupDialogs[which];
	if ( not info ) then
		return nil;
	end

	local text = _G[dialog:GetName().."Text"];
	local editBox = _G[dialog:GetName().."EditBox"];
	local button1 = _G[dialog:GetName().."Button1"];
	local extraButton = dialog.extraButton;

	local maxHeightSoFar, maxWidthSoFar = (dialog.maxHeightSoFar or 0), (dialog.maxWidthSoFar or 0);
	local width = 320;

	if ( info.verticalButtonLayout ) then
		width = width + 30;
	else
		if ( dialog.numButtons == 4 ) then
			width = 574;
		elseif ( dialog.numButtons == 3 ) then
			width = 440;
		elseif (info.showAlert or info.showAlertGear or info.closeButton or info.wide) then
			-- Widen
			width = 420;
		elseif ( info.editBoxWidth and info.editBoxWidth > 260 ) then
			width = width + (info.editBoxWidth - 260);
		elseif ( which == "HELP_TICKET" ) then
			width = 350;
		elseif ( which == "GUILD_IMPEACH" ) then
			width = 375;
		end
	end
	if ( dialog.insertedFrame ) then
		width = max(width, dialog.insertedFrame:GetWidth());
	end
	if ( width > maxWidthSoFar )  then
		dialog:SetWidth(width);
		dialog.maxWidthSoFar = width;
	end

	local height = 32 + text:GetHeight() + 2;
	if ( info.extraButton ) then
		height = height + 40 + extraButton:GetHeight();
	end
	if ( not info.nobuttons ) then
		height = height + 6 + button1:GetHeight();
	end
	if ( info.hasEditBox ) then
		height = height + 8 + editBox:GetHeight();
	elseif ( info.hasMoneyFrame ) then
		height = height + 16;
	elseif ( info.hasMoneyInputFrame ) then
		height = height + 22;
	end
	if ( dialog.insertedFrame ) then
		height = height + dialog.insertedFrame:GetHeight();
	end
	if ( info.hasItemFrame ) then
		height = height + 64;
	end
	if ( dialog.SubText:IsShown() ) then
		height = height + dialog.SubText:GetHeight() + 8;
	end

	if ( info.verticalButtonLayout ) then
		height = height + 16 + (26 * (dialog.numButtons - 1));
	end

	if ( height > maxHeightSoFar ) then
		dialog:SetHeight(height);
		dialog.maxHeightSoFar = height;
	end
end

local tempButtonLocs = {};	--So we don't make a new table each time.
function SafeStaticPopup_Show(which, text_arg1, text_arg2, data, insertedFrame)
	local info = SafeStaticPopupDialogs[which];
	if ( not info ) then
		return nil;
	end

	if ( info.OnAccept and info.OnButton1 ) then
		error("Dialog "..which.. " cannot have both OnAccept and OnButton1");
	end
	if ( info.OnCancel and info.OnButton2 ) then
		error("Dialog "..which.. " cannot have both OnCancel and OnButton2");
	end

	if ( UnitIsDeadOrGhost("player") and not info.whileDead ) then
		if ( info.OnCancel ) then
			info.OnCancel();
		end
		return nil;
	end

	if ( InCinematic() and not info.interruptCinematic ) then
		if ( info.OnCancel ) then
			info.OnCancel();
		end
		return nil;
	end

	if ( info.exclusive ) then
		SafeStaticPopup_HideExclusive();
	end

	if ( info.cancels ) then
		for index = 1, SAFESTATICPOPUP_NUMDIALOGS, 1 do
			local frame = _G["SafeStaticPopup"..index];
			if ( frame:IsShown() and (frame.which == info.cancels) ) then
				frame:Hide();
				local OnCancel = SafeStaticPopupDialogs[frame.which].OnCancel;
				if ( OnCancel ) then
					OnCancel(frame, frame.data, "override");
				end
			end
		end
	end

	-- Pick a free dialog to use
	local dialog = nil;
	-- Find an open dialog of the requested type
	dialog = SafeStaticPopup_FindVisible(which, data);
	if ( dialog ) then
		if ( not info.noCancelOnReuse ) then
			local OnCancel = info.OnCancel;
			if ( OnCancel ) then
				OnCancel(dialog, dialog.data, "override");
			end
		end
		dialog:Hide();
	end
	if ( not dialog ) then
		-- Find a free dialog
		local index = 1;
		if ( info.preferredIndex ) then
			index = info.preferredIndex;
		end
		for i = index, SAFESTATICPOPUP_NUMDIALOGS do
			local frame = _G["SafeStaticPopup"..i];
			if ( not frame:IsShown() ) then
				dialog = frame;
				break;
			end
		end

		--If dialog not found and there's a preferredIndex then try to find an available frame before the preferredIndex
		if ( not dialog and info.preferredIndex ) then
			for i = 1, info.preferredIndex do
				local frame = _G["SafeStaticPopup"..i];
				if ( not frame:IsShown() ) then
					dialog = frame;
					break;
				end
			end
		end
	end
	if ( not dialog ) then
		if ( info.OnCancel ) then
			info.OnCancel();
		end
		return nil;
	end

	dialog.CoverFrame:SetShown(info.fullScreenCover);

	dialog.maxHeightSoFar, dialog.maxWidthSoFar = 0, 0;
	local bottomSpace = info.extraButton ~= nil and (dialog.extraButton:GetHeight() + 60) or 16;

	-- Set the text of the dialog
	local text = _G[dialog:GetName().."Text"];
	text:SetFormattedText(info.text, text_arg1, text_arg2);
	text.text_arg1 = text_arg1;
	text.text_arg2 = text_arg2;

	-- Show or hide the close button
	if ( info.closeButton ) then
		local closeButton = _G[dialog:GetName().."CloseButton"];
		if ( info.closeButtonIsHide ) then
			closeButton:SetNormalTexture("Interface\\Buttons\\UI-Panel-HideButton-Up");
			closeButton:SetPushedTexture("Interface\\Buttons\\UI-Panel-HideButton-Down");
		else
			closeButton:SetNormalTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up");
			closeButton:SetPushedTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Down");
		end
		closeButton:Show();
	else
		_G[dialog:GetName().."CloseButton"]:Hide();
	end

	-- Set the editbox of the dialog
	local editBox = _G[dialog:GetName().."EditBox"];
	if ( info.hasEditBox ) then
		editBox:Show();

		editBox.Instructions:SetText(info.editBoxInstructions or "");

		if ( info.maxLetters ) then
			editBox:SetMaxLetters(info.maxLetters);
			editBox:SetCountInvisibleLetters(info.countInvisibleLetters);
		end
		if ( info.maxBytes ) then
			editBox:SetMaxBytes(info.maxBytes);
		end
		editBox:SetText("");
		if ( info.editBoxWidth ) then
			editBox:SetWidth(info.editBoxWidth);
		else
			editBox:SetWidth(130);
		end

		editBox:ClearAllPoints();
		editBox:SetPoint("BOTTOM", 0, 29 + bottomSpace);
	else
		editBox:Hide();
	end

	-- Show or hide money frame
	if ( info.hasMoneyFrame ) then
		_G[dialog:GetName().."MoneyFrame"]:Show();
		_G[dialog:GetName().."MoneyInputFrame"]:Hide();
	elseif ( info.hasMoneyInputFrame ) then
		local moneyInputFrame = _G[dialog:GetName().."MoneyInputFrame"];
		moneyInputFrame:Show();
		_G[dialog:GetName().."MoneyFrame"]:Hide();
		-- Set OnEnterPress for money input frames
		if ( info.EditBoxOnEnterPressed ) then
			moneyInputFrame.gold:SetScript("OnEnterPressed", SafeStaticPopup_EditBoxOnEnterPressed);
			moneyInputFrame.silver:SetScript("OnEnterPressed", SafeStaticPopup_EditBoxOnEnterPressed);
			moneyInputFrame.copper:SetScript("OnEnterPressed", SafeStaticPopup_EditBoxOnEnterPressed);
		else
			moneyInputFrame.gold:SetScript("OnEnterPressed", nil);
			moneyInputFrame.silver:SetScript("OnEnterPressed", nil);
			moneyInputFrame.copper:SetScript("OnEnterPressed", nil);
		end
	else
		_G[dialog:GetName().."MoneyFrame"]:Hide();
		_G[dialog:GetName().."MoneyInputFrame"]:Hide();
	end

	dialog.ItemFrame.itemID = nil;
	-- Show or hide item button
	if ( info.hasItemFrame ) then
		dialog.ItemFrame:Show();
		if ( data and type(data) == "table" ) then
			if ( data.useLinkForItemInfo ) then
				SafeStaticPopupItemFrame_RetrieveInfo(dialog.ItemFrame, data);
			end
			SafeStaticPopupItemFrame_DisplayInfo(dialog.ItemFrame, data.link, data.name, data.color, data.texture, data.count);
		end
	else
		dialog.ItemFrame:Hide();
	end

	-- Set the miscellaneous variables for the dialog
	dialog.which = which;
	dialog.timeleft = info.timeout or 0;
	dialog.hideOnEscape = info.hideOnEscape;
	dialog.exclusive = info.exclusive;
	dialog.enterClicksFirstButton = info.enterClicksFirstButton;
	dialog.insertedFrame = insertedFrame;
	if ( info.subText ) then
		dialog.SubText:SetText(info.subText);
		dialog.SubText:Show();
	else
		dialog.SubText:Hide();
	end

	if ( insertedFrame ) then
		insertedFrame:SetParent(dialog);
		insertedFrame:ClearAllPoints();
		if ( dialog.SubText:IsShown() ) then
			insertedFrame:SetPoint("TOP", dialog.SubText, "BOTTOM");
		else
			insertedFrame:SetPoint("TOP", text, "BOTTOM");
		end
		insertedFrame:Show();
		_G[dialog:GetName().."MoneyFrame"]:SetPoint("TOP", insertedFrame, "BOTTOM");
		_G[dialog:GetName().."MoneyInputFrame"]:SetPoint("TOP", insertedFrame, "BOTTOM");
	elseif ( dialog.SubText:IsShown() ) then
		_G[dialog:GetName().."MoneyFrame"]:SetPoint("TOP", dialog.SubText, "BOTTOM", 0, -5);
		_G[dialog:GetName().."MoneyInputFrame"]:SetPoint("TOP", dialog.SubText, "BOTTOM", 0, -5);
	else
		_G[dialog:GetName().."MoneyFrame"]:SetPoint("TOP", dialog.text, "BOTTOM", 0, -5);
		_G[dialog:GetName().."MoneyInputFrame"]:SetPoint("TOP", dialog.text, "BOTTOM", 0, -5);
	end
	-- Clear out data
	dialog.data = data;

	-- Set the buttons of the dialog
	local button1 = _G[dialog:GetName().."Button1"];
	local button2 = _G[dialog:GetName().."Button2"];
	local button3 = _G[dialog:GetName().."Button3"];
	local button4 = _G[dialog:GetName().."Button4"];

	do	--If there is any recursion in this block, we may get errors (tempButtonLocs is static). If you have to recurse, we'll have to create a new table each time.
		assert(#tempButtonLocs == 0);	--If this fails, we're recursing. (See the table.wipe at the end of the block)

		tinsert(tempButtonLocs, button1);
		tinsert(tempButtonLocs, button2);
		tinsert(tempButtonLocs, button3);
		tinsert(tempButtonLocs, button4);

		for i=#tempButtonLocs, 1, -1 do
			--Do this stuff before we move it. (This is why we go back-to-front)
			tempButtonLocs[i]:SetText(info["button"..i]);
			tempButtonLocs[i]:Hide();
			tempButtonLocs[i]:ClearAllPoints();
			tempButtonLocs[i].PulseAnim:Stop();
			--Now we possibly remove it.
			if ( not (info["button"..i] and ( not info["DisplayButton"..i] or info["DisplayButton"..i](dialog))) ) then
				tremove(tempButtonLocs, i);
			end
		end

		local numButtons = #tempButtonLocs;
		--Save off the number of buttons.
		dialog.numButtons = numButtons;

		if numButtons > 0 then
			tempButtonLocs[1]:ClearAllPoints();
			if ( info.verticalButtonLayout ) then
				tempButtonLocs[1]:SetPoint("TOP", dialog.text, "BOTTOM", 0, -16);
			else
				if ( numButtons == 4 ) then
					tempButtonLocs[1]:SetPoint("BOTTOMRIGHT", dialog, "BOTTOM", -139, bottomSpace);
				elseif ( numButtons == 3 ) then
					tempButtonLocs[1]:SetPoint("BOTTOMRIGHT", dialog, "BOTTOM", -72, bottomSpace);
				elseif ( numButtons == 2 ) then
					tempButtonLocs[1]:SetPoint("BOTTOMRIGHT", dialog, "BOTTOM", -6, bottomSpace);
				elseif ( numButtons == 1 ) then
					tempButtonLocs[1]:SetPoint("BOTTOM", dialog, "BOTTOM", 0, bottomSpace);
				end
			end
		end

		for i=1, numButtons do
			if ( i > 1 ) then
				tempButtonLocs[i]:ClearAllPoints();
				if info.verticalButtonLayout then
					tempButtonLocs[i]:SetPoint("TOP", tempButtonLocs[i-1], "BOTTOM", 0, -6);
				else
					tempButtonLocs[i]:SetPoint("LEFT", tempButtonLocs[i-1], "RIGHT", 13, 0);
				end
			end

			local width = tempButtonLocs[i]:GetTextWidth();
			if ( width > 110 ) then
				tempButtonLocs[i]:SetWidth(width + 20);
			else
				tempButtonLocs[i]:SetWidth(120);
			end
			if (info["button"..i.."Pulse"]) then
				tempButtonLocs[i].PulseAnim:Play();
			end
			tempButtonLocs[i]:Enable();
			tempButtonLocs[i]:Show();
		end

		table.wipe(tempButtonLocs);
	end

	if info.extraButton then
		local extraButton = dialog.extraButton;
		extraButton:Show();
		extraButton:SetPoint("BOTTOM", dialog, "BOTTOM", 0, 22);
		extraButton:SetText(info.extraButton);
		--widen if too small, but reset to 128 otherwise
		local width = 128
		local padding = 40;
		local textWidth = extraButton:GetTextWidth() + padding;
		width = math.max(width, textWidth)
		extraButton:SetWidth(width)

		dialog.Separator:Show();
	else
		dialog.extraButton:Hide();
		dialog.Separator:Hide();
	end

	-- Show or hide the alert icon
	local alertIcon = _G[dialog:GetName().."AlertIcon"];
	if ( info.showAlert ) then
		alertIcon:SetTexture(SAFESTATICPOPUP_TEXTURE_ALERT);
		if ( button3:IsShown() )then
			alertIcon:SetPoint("LEFT", 24, 10);
		else
			alertIcon:SetPoint("LEFT", 24, 0);
		end
		alertIcon:Show();
	elseif ( info.showAlertGear ) then
		alertIcon:SetTexture(SAFESTATICPOPUP_TEXTURE_ALERTGEAR);
		if ( button3:IsShown() )then
			alertIcon:SetPoint("LEFT", 24, 0);
		else
			alertIcon:SetPoint("LEFT", 24, 0);
		end
		alertIcon:Show();
	else
		alertIcon:SetTexture();
		alertIcon:Hide();
	end

	if ( info.StartDelay ) then
		dialog.startDelay = info.StartDelay(dialog);
		if (not dialog.startDelay or dialog.startDelay <= 0) then
			button1:Enable();
		else
			button1:Disable();
		end
	else
		dialog.startDelay = nil;
		button1:Enable();
	end

	editBox:SetSecureText(info.editBoxSecureText);
	editBox.hasAutoComplete = info.autoCompleteSource ~= nil;
	if ( editBox.hasAutoComplete ) then
		AutoCompleteEditBox_SetAutoCompleteSource(editBox, info.autoCompleteSource, unpack(info.autoCompleteArgs));
	else
		AutoCompleteEditBox_SetAutoCompleteSource(editBox, nil);
	end

	-- Finally size and show the dialog
	SafeStaticPopup_SetUpPosition(dialog);
	dialog:Show();

	SafeStaticPopup_Resize(dialog, which);

	if ( info.sound ) then
		PlaySound(info.sound);
	end

	return dialog;
end

function SafeStaticPopup_Hide(which, data)
	for index = 1, SAFESTATICPOPUP_NUMDIALOGS, 1 do
		local dialog = _G["SafeStaticPopup"..index];
		if ( dialog:IsShown() and (dialog.which == which) and (not data or (data == dialog.data)) ) then
			dialog:Hide();
		end
	end
end

function SafeStaticPopup_OnUpdate(dialog, elapsed)
	if ( dialog.timeleft > 0 ) then
		local which = dialog.which;
		local timeleft = dialog.timeleft - elapsed;
		if ( timeleft <= 0 ) then
			if ( not SafeStaticPopupDialogs[which].timeoutInformationalOnly ) then
				dialog.timeleft = 0;
				local OnCancel = SafeStaticPopupDialogs[which].OnCancel;
				if ( OnCancel ) then
					OnCancel(dialog, dialog.data, "timeout");
				end
				dialog:Hide();
			end
			return;
		end
		dialog.timeleft = timeleft;

		if ( (which == "DEATH") or
		     (which == "CAMP")  or
			 (which == "QUIT") or
			 (which == "DUEL_OUTOFBOUNDS") or
			 (which == "INSTANCE_BOOT") or
			 (which == "GARRISON_BOOT") or
			 (which == "CONFIRM_SUMMON") or
			 (which == "CONFIRM_SUMMON_SCENARIO") or
			 (which == "CONFIRM_SUMMON_STARTING_AREA") or
			 (which == "BFMGR_INVITED_TO_ENTER") or
			 (which == "AREA_SPIRIT_HEAL") or
			 (which == "SPELL_CONFIRMATION_PROMPT") ) then
			local text = _G[dialog:GetName().."Text"];
			timeleft = ceil(timeleft);
			if ( (which == "INSTANCE_BOOT") or (which == "GARRISON_BOOT") ) then
				if ( timeleft < 60 ) then
					text:SetFormattedText(SafeStaticPopupDialogs[which].text, timeleft, SECONDS);
				else
					text:SetFormattedText(SafeStaticPopupDialogs[which].text, ceil(timeleft / 60), MINUTES);
				end
			elseif ( which == "CONFIRM_SUMMON" or which == "CONFIRM_SUMMON_SCENARIO" or which == "CONFIRM_SUMMON_STARTING_AREA" ) then
				if ( timeleft < 60 ) then
					text:SetFormattedText(SafeStaticPopupDialogs[which].text, C_SummonInfo.GetSummonConfirmSummoner() or "", C_SummonInfo.GetSummonConfirmAreaName(), timeleft, SECONDS);
				else
					text:SetFormattedText(SafeStaticPopupDialogs[which].text, C_SummonInfo.GetSummonConfirmSummoner() or "", C_SummonInfo.GetSummonConfirmAreaName(), ceil(timeleft / 60), MINUTES);
				end
			elseif ( which == "BFMGR_INVITED_TO_ENTER") then
				if ( timeleft < 60 ) then
					text:SetFormattedText(SafeStaticPopupDialogs[which].text, text.text_arg1, timeleft, SECONDS);
				else
					text:SetFormattedText(SafeStaticPopupDialogs[which].text, text.text_arg1, ceil(timeleft / 60), MINUTES);
				end
			elseif ( which == "SPELL_CONFIRMATION_PROMPT") then
				local time = "";
				if ( timeleft < 60 ) then
					text:SetFormattedText(ERR_SPELL_FAILED_S, timeleft, SECONDS);
				else
					text:SetFormattedText(ERR_SPELL_FAILED_S, ceil(timeleft / 60), MINUTES);
				end
				text:SetText(SafeStaticPopupDialogs[which].text .. " " ..TIME_REMAINING .. text:GetText());
			else
				if ( timeleft < 60 ) then
					text:SetFormattedText(SafeStaticPopupDialogs[which].text, timeleft, SECONDS);
				else
					text:SetFormattedText(SafeStaticPopupDialogs[which].text, ceil(timeleft / 60), MINUTES);
				end
			end
			SafeStaticPopup_Resize(dialog, which);
		end
	end
	if ( dialog.startDelay ) then
		local which = dialog.which;
		local timeleft = dialog.startDelay - elapsed;
		if ( timeleft <= 0 ) then
			dialog.startDelay = nil;
			local text = _G[dialog:GetName().."Text"];
			text:SetFormattedText(SafeStaticPopupDialogs[which].text, text.text_arg1, text.text_arg2);
			local button1 = _G[dialog:GetName().."Button1"];
			button1:Enable();
			SafeStaticPopup_Resize(dialog, which);
			return;
		end
		dialog.startDelay = timeleft;

		if ( which == "RECOVER_CORPSE" or (which == "RESURRECT") or (which == "RESURRECT_NO_SICKNESS") ) then
			local text = _G[dialog:GetName().."Text"];
			timeleft = ceil(timeleft);
			if ( (which == "RESURRECT") or (which == "RESURRECT_NO_SICKNESS") ) then
				if ( timeleft < 60 ) then
					text:SetFormattedText(SafeStaticPopupDialogs[which].delayText, text.text_arg1, timeleft, SECONDS);
				else
					text:SetFormattedText(SafeStaticPopupDialogs[which].delayText, text.text_arg1, ceil(timeleft / 60), MINUTES);
				end
			else
				if ( timeleft < 60 ) then
					text:SetFormattedText(SafeStaticPopupDialogs[which].delayText, timeleft, SECONDS);
				else
					text:SetFormattedText(SafeStaticPopupDialogs[which].delayText, ceil(timeleft / 60), MINUTES);
				end
			end
			SafeStaticPopup_Resize(dialog, which);
		end
	end

	local onUpdate = SafeStaticPopupDialogs[dialog.which].OnUpdate;
	if ( onUpdate ) then
		onUpdate(dialog, elapsed);
	end
end

function SafeStaticPopup_EditBoxOnEnterPressed(self)
	local EditBoxOnEnterPressed, which, dialog;
	local parent = self:GetParent();
	if ( parent.which ) then
		which = parent.which;
		dialog = parent;
	elseif ( parent:GetParent().which ) then
		-- This is needed if this is a money input frame since it's nested deeper than a normal edit box
		which = parent:GetParent().which;
		dialog = parent:GetParent();
	end
	if ( not self.hasAutoComplete or not AutoCompleteEditBox_OnEnterPressed(self) ) then
		EditBoxOnEnterPressed = SafeStaticPopupDialogs[which].EditBoxOnEnterPressed;
		if ( EditBoxOnEnterPressed ) then
			EditBoxOnEnterPressed(self, dialog.data);
		end
	end
end

function SafeStaticPopup_EditBoxOnEscapePressed(self)
	local EditBoxOnEscapePressed = SafeStaticPopupDialogs[self:GetParent().which].EditBoxOnEscapePressed;
	if ( EditBoxOnEscapePressed ) then
		EditBoxOnEscapePressed(self, self:GetParent().data);
	end
end

function SafeStaticPopup_EditBoxOnTextChanged(self, userInput)
	if ( not self.hasAutoComplete or not AutoCompleteEditBox_OnTextChanged(self, userInput) ) then
		local EditBoxOnTextChanged = SafeStaticPopupDialogs[self:GetParent().which].EditBoxOnTextChanged;
		if ( EditBoxOnTextChanged ) then
			EditBoxOnTextChanged(self, self:GetParent().data);
		end
	end
	self.Instructions:SetShown(self:GetText() == "");
end

function SafeStaticPopup_OnLoad(self)
	local name = self:GetName();
	self.button1 = _G[name .. "Button1"];
	self.button2 = _G[name .. "Button2"];
	self.button3 = _G[name .. "Button3"];
	self.text = _G[name .. "Text"];
	self.icon = _G[name .. "AlertIcon"];
	self.moneyInputFrame = _G[name .. "MoneyInputFrame"];
	self:RegisterEvent("DISPLAY_SIZE_CHANGED");
end

function SafeStaticPopup_OnShow(self)
	PlaySound(SOUNDKIT.IG_MAINMENU_OPEN);

	local dialog = SafeStaticPopupDialogs[self.which];
	local OnShow = dialog.OnShow;

	if ( OnShow ) then
		OnShow(self, self.data);
	end
	if ( dialog.hasMoneyInputFrame ) then
		_G[self:GetName().."MoneyInputFrameGold"]:SetFocus();
	end
	if ( dialog.enterClicksFirstButton ) then
		self:SetScript("OnKeyDown", SafeStaticPopup_OnKeyDown);
	end
end

function SafeStaticPopup_OnHide(self)
	PlaySound(SOUNDKIT.IG_MAINMENU_CLOSE);

	SafeStaticPopup_CollapseTable();

	local dialog = SafeStaticPopupDialogs[self.which];
	local OnHide = dialog.OnHide;
	if ( OnHide ) then
		OnHide(self, self.data);
	end
	self.extraFrame:Hide();
	if ( dialog.enterClicksFirstButton ) then
		self:SetScript("OnKeyDown", nil);
	end
	if ( self.insertedFrame ) then
		self.insertedFrame:Hide();
		self.insertedFrame:SetParent(nil);
		local text = _G[self:GetName().."Text"];
		_G[self:GetName().."MoneyFrame"]:SetPoint("TOP", text, "BOTTOM", 0, -5);
		_G[self:GetName().."MoneyInputFrame"]:SetPoint("TOP", text, "BOTTOM", 0, -5);
	end
end

local function SafeStaticPopup_CallInfoHandler(dialog, handlerName, ...)
	if ( dialog:IsShown() ) then
		local which = dialog.which;
		local info = SafeStaticPopupDialogs[which];
		if ( info ) then
			local handler = info[handlerName];
			if ( handler ) then
				handler(dialog, ...);
			end
		end
	end
end

function SafeStaticPopup_OnHyperlinkClick(self, ...)
	SafeStaticPopup_CallInfoHandler(self, "OnHyperlinkClick", ...);
end

function SafeStaticPopup_OnHyperlinkEnter(self, ...)
	SafeStaticPopup_CallInfoHandler(self, "OnHyperlinkEnter", ...);
end

function SafeStaticPopup_OnHyperlinkLeave(self, ...)
	SafeStaticPopup_CallInfoHandler(self, "OnHyperlinkLeave", ...);
end

function SafeStaticPopup_OnClick(dialog, index)
	if ( not dialog:IsShown() ) then
		return;
	end
	local which = dialog.which;
	local info = SafeStaticPopupDialogs[which];
	if ( not info ) then
		return nil;
	end

	if ( which == "DEATH" or which == "CLASS_TRIAL_CHOOSE_BOOST_TYPE" ) then
		local func;
		if ( index == 1 ) then
			func = info.OnAccept or info.OnButton1;
		elseif ( index == 2 ) then
			func = info.OnCancel or info.OnButton2;
		elseif ( index == 3 ) then
			func = info.OnButton3;
		elseif ( index == 4 ) then
			func = info.OnButton4;
		elseif ( index == 5 ) then
			func = info.OnExtraButton;
		end

		if ( func ) then
			local keepOpen = func(dialog, dialog.data, "clicked");
			if ( not keepOpen and which == dialog.which ) then
				dialog:Hide();
			end
		end
	else
		-- Keeping this temporarily for backward compatibility
		local hide = true;
		if ( index == 1 ) then
			local OnAccept = info.OnAccept or info.OnButton1;
			if ( OnAccept ) then
				hide = not OnAccept(dialog, dialog.data, dialog.data2);
			end
		elseif ( index == 3 ) then
			local OnAlt = info.OnAlt;
			if ( OnAlt ) then
				OnAlt(dialog, dialog.data, "clicked");
			end
		elseif ( index == 5 ) then
			local OnExtraButton = info.OnExtraButton;
			if ( OnExtraButton ) then
				OnExtraButton(dialog, dialog.data, dialog.data2);
			end
		else
			local OnCancel = info.OnCancel;
			if ( OnCancel ) then
				hide = not OnCancel(dialog, dialog.data, "clicked");
			end
		end

		if ( hide and (which == dialog.which) and ( index ~= 3 or not info.noCloseOnAlt) ) then
			-- can dialog.which change inside one of the On* functions???
			dialog:Hide();
		end
	end
end

function SafeStaticPopup_OnKeyDown(self, key)
	-- previously, SafeStaticPopup_EscapePressed() captured the escape key for dialogs, but now we need
	-- to catch it here
	if ( GetBindingFromClick(key) == "TOGGLEGAMEMENU" ) then
		return SafeStaticPopup_EscapePressed();
	elseif ( GetBindingFromClick(key) == "SCREENSHOT" ) then
		RunBinding("SCREENSHOT");
		return;
	end

	local dialog = SafeStaticPopupDialogs[self.which];
	if ( dialog ) then
		if ( key == "ENTER" and dialog.enterClicksFirstButton ) then
			local frameName = self:GetName();
			local button;
			local i = 1;
			while ( true ) do
				button = _G[frameName.."Button"..i];
				if ( button ) then
					if ( button:IsShown() ) then
						SafeStaticPopup_OnClick(self, i);
						return;
					end
					i = i + 1;
				else
					break;
				end
			end
		end
	end
end

function SafeStaticPopup_EscapePressed()
	local closed = nil;
	for _, frame in pairs(SafeStaticPopup_DisplayedFrames) do
		if( frame:IsShown() and frame.hideOnEscape ) then
			local standardDialog = SafeStaticPopupDialogs[frame.which];
			if ( standardDialog ) then
				local OnCancel = standardDialog.OnCancel;
				local noCancelOnEscape = standardDialog.noCancelOnEscape;
				if ( OnCancel and not noCancelOnEscape) then
					OnCancel(frame, frame.data, "clicked");
				end
				frame:Hide();
			else
				SafeStaticPopupSpecial_Hide(frame);
			end
			closed = 1;
		end
	end
	return closed;
end

function SafeStaticPopup_SetUpPosition(dialog)
	if ( not tContains(SafeStaticPopup_DisplayedFrames, dialog) ) then
		SafeStaticPopup_SetUpAnchor(dialog, #SafeStaticPopup_DisplayedFrames + 1);
		tinsert(SafeStaticPopup_DisplayedFrames, dialog);
	end
end

function SafeStaticPopup_SetUpAnchor(dialog, idx)
	local lastFrame = SafeStaticPopup_DisplayedFrames[idx - 1];
	if ( lastFrame ) then
		dialog:SetPoint("TOP", lastFrame, "BOTTOM", 0, 0);
	else
		dialog:SetPoint("TOP", UIParent, "TOP", 0, -135);
	end
end

function SafeStaticPopup_CollapseTable()
	local displayedFrames = SafeStaticPopup_DisplayedFrames;
	local index = #displayedFrames;
	while ( ( index >= 1 ) and ( not displayedFrames[index]:IsShown() ) ) do
		tremove(displayedFrames, index);
		index = index - 1;
	end
end

function SafeStaticPopupSpecial_Show(frame)
	if ( frame.exclusive ) then
		SafeStaticPopup_HideExclusive();
	end
	SafeStaticPopup_SetUpPosition(frame);
	frame:Show();
end

function SafeStaticPopupSpecial_Hide(frame)
	frame:Hide();
	SafeStaticPopup_CollapseTable();
end

function SafeStaticPopupSpecial_Toggle(frame)
	if frame:IsShown() then
		SafeStaticPopupSpecial_Hide(frame);
	else
		SafeStaticPopupSpecial_Show(frame);
	end
end

--Note that things will look sub-fantastic if toActivate is bigger than toReplace
function SafeStaticPopupSpecial_Replace(toActivate, toReplace)
	local idx = nil;
	for i=1, #SafeStaticPopup_DisplayedFrames do
		if ( SafeStaticPopup_DisplayedFrames[i] == toReplace ) then
			idx = i;
			break;
		end
	end

	if ( idx ) then
		SafeStaticPopup_DisplayedFrames[idx] = toActivate;
		SafeStaticPopup_SetUpAnchor(toActivate, idx);

		toReplace:Hide();
		toActivate:Show();
		return true;
	end

	return false;
end

--Used to figure out if we can resize a frame
function SafeStaticPopup_IsLastDisplayedFrame(frame)
	for i=#SafeStaticPopup_DisplayedFrames, 1, -1 do
		local popup = SafeStaticPopup_DisplayedFrames[i];
		if ( popup:IsShown() ) then
			return frame == popup
		end
	end
	return false;
end

function SafeStaticPopup_OnEvent(self)
	self.maxHeightSoFar = 0;
	SafeStaticPopup_Resize(self, self.which);
end

function SafeStaticPopup_HideExclusive()
	for _, frame in pairs(SafeStaticPopup_DisplayedFrames) do
		if ( frame:IsShown() and frame.exclusive ) then
			local standardDialog = SafeStaticPopupDialogs[frame.which];
			if ( standardDialog ) then
				frame:Hide();
				local OnCancel = standardDialog.OnCancel;
				if ( OnCancel ) then
					OnCancel(frame, frame.data, "override");
				end
			else
				SafeStaticPopupSpecial_Hide(frame);
			end
			break;
		end
	end
end

function SafeStaticPopupItemFrame_OnEvent(self, event, ...)
	if ( event == "GET_ITEM_INFO_RECEIVED" ) then
		local itemID = ...;
		if ( itemID == self.itemID ) then
			local data = self:GetParent().data;
			SafeStaticPopupItemFrame_RetrieveInfo(self, data);
			SafeStaticPopupItemFrame_DisplayInfo(self, data.link, data.name, data.color, data.texture, data.count);
		end
	end
end

function SafeStaticPopupItemFrame_RetrieveInfo(self, data)
	local itemName, _, itemQuality, _, _, _, _, _, _, texture = GetItemInfo(data.link);
	if ( itemName ) then
		data.name = itemName;
		local r, g, b = GetItemQualityColor(itemQuality);
		data.color = {r, g, b, 1};
		data.texture = texture;
		self.itemID = nil;
	else
		local itemID, _, _, _, texture = GetItemInfoInstant(data.link);
		data.name = RETRIEVING_ITEM_INFO;
		data.color = {RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b, 1};
		data.texture = texture;
		self.itemID = itemID;
	end
end

function SafeStaticPopupItemFrame_DisplayInfo(self, link, name, color, texture, count)
	self.link = link;
	_G[self:GetName().."IconTexture"]:SetTexture(texture);
	local nameText = _G[self:GetName().."Text"];
	nameText:SetTextColor(unpack(color or {1, 1, 1, 1}));
	nameText:SetText(name);

	if link then
		local quality = select(3, GetItemInfo(link));
		SetItemButtonQuality(self, quality, link);
	end

	if ( count and count > 1 ) then
		_G[self:GetName().."Count"]:SetText(count);
		_G[self:GetName().."Count"]:Show();
	else
		_G[self:GetName().."Count"]:Hide();
	end
end
